# (c) Copyright IBM Corp. 2010, 2018. All Rights Reserved.

import pkg_resources
try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    __version__ = None

from .actions_component import ResilientComponent
from .action_message import ActionMessageBase, ActionMessage, \
    FunctionMessage, FunctionResult, FunctionError, \
    StatusMessage, BaseFunctionError
from .decorators import function, inbound_app, app_function, handler, required_field, required_action_field, defer, debounce
from .actions_test_component import SubmitTestAction, SubmitTestFunction, SubmitTestInboundApp
from .app_function_component import AppFunctionComponent
from .helpers import is_this_a_selftest
